#include <cstdio>
#include <algorithm>
#include <vector>
#include <set>
#include <ext/slist>
#include <stdexcept>


#include "graph.hpp"
#include "kmer.hpp"
#include "fastq.hpp"
#include "kseq.h"

 

size_t help_deg(size_t link, bool fw);

struct PairPairSortSecond {
  bool operator()(const pair<int, pair<int,int> > &l, const pair<int, pair<int,int> > &r) {
    return (l.first < r.first) || (l.first == r.first && (l.second.second < r.second.second || ((l.second.second == r.second.second ) && (l.second.first < r.second.first))));
  }
};

uint8_t char_to_ind(char c) {
  switch (c) {
  case 'A': return 0; break;
  case 'C': return 1; break;
  case 'G': return 2; break;
  case 'T': return 3; break;
  }
  return 0xFF;
}

char ind_to_char(size_t i) {
  switch (i) {
  case 0: return 'A';
  case 1: return 'C';
  case 2: return 'G';
  case 3: return 'T';
  }
  return '\0';
}

int mismatch(string &a, string &b, int maxMismatch) {
  return 0;
}


uint8_t revbits(uint8_t b) {
  return (b * 0x0202020202ULL & 0x010884422010ULL) % 1023;
}

size_t linkmask(char c, bool fw) {
  size_t pos = char_to_ind(c);

  if (!fw && pos != 0xFF) {
    pos += 4;
  }
  if (pos != 0xFF) {
    return (1 << pos);
  } else {
    return 0;
  }
}

size_t link_to_ind(size_t link) {
  switch (link) {
  case 1: return 0;
  case 2: return 1;
  case 4: return 2;
  case 8: return 3;
  default: return 0xFF;
  }
}

GraphBuilder::~GraphBuilder() {
  if (f != NULL) {
    delete f;
    f = NULL;
  }
}

void GraphBuilder::processFastqFile(FastqFile &F) {
  size_t k = Kmer::k;
  size_t kmers_per_read = (k>0) ? (76-Kmer::k)*2 : 100;
  
  //double fudge_factor = 0.40;
  //size_t fsize = exp_size * kmers_per_read / 2;

  if (G.kmap.size() == 0) {
    G.kmap.resize((size_t) 500000000ULL); // guess

    printf("size: %zu, bucket_count: %zu\n",G.kmap.size(),G.kmap.bucket_count());
  }

  // build bloom filter
  if (f == NULL) {
    f = new bloom_filter(exp_size * kmers_per_read, (size_t) 4, (unsigned long) time(NULL));
  }


  char name[1024],s[1024];
  size_t name_len,len;
  hmap_t::iterator it,it2;
  pair<hmap_t::iterator, bool> ref,ref2;
  size_t n_read = 0;
  while (F.read_next(name, &name_len,s,&len,NULL)>=0) {
    Kmer km(s);
    for (size_t i = 0; i <= len-k; ++i) {
      if (i > 0) {
	km = km.forwardBase(s[i+k-1]);
      }
      Kmer tw = km.twin();
      Kmer rep = (km < tw) ? km : tw;
      if (f->contains(rep)) {
	G.addKmer(km);
      } else {
	f->insert(rep);
      }
    }
    ++n_read;

    if (n_read % 1000000 == 0) {
      printf("processed %zu reads\n",n_read);
    }
  }
  
  F.reopen();
  
  while (F.read_next(name,&name_len,s,&len,NULL)>=0) {
    //printf("len:%d, k: %d\n",len,k);
    //printf("%s\n",s);
    
    Kmer km(s);
    
    for (size_t i =0; i <= len-k; i++) {
      if (i > 0) {
	km = km.forwardBase(s[i+k-1]);
      }

      if (G.kmap.find(km) != G.kmap.end()) {
	char b_base = '\0';
	char f_base = '\0';

	if (i > 0) { b_base = s[i-1];}
	if (i < len-k) { f_base = s[i+k];}

	G.addKmer(km,f_base,b_base,true);
      }
    }
  }
  F.close();

  Kmer km_del;
  km_del.set_deleted();
  G.kmap.set_deleted_key(km_del);


  for (it = G.kmap.begin(); it != G.kmap.end(); ) {
    if (it->second.cov <= 1) {
      hmap_t::iterator del(it);
      ++it;
      G.removeKmer(del);
    } else {
      ++it;
    }
  }

  // fix all links
  //char tmp[1024];
  for (it = G.kmap.begin(); it != G.kmap.end(); ++it ) {
    for (int j = 0; j < 8; ++j) {
      if (it->second.link & (1<<j)) {
	Kmer nb = it->first.getLink(j);
	if (G.kmap.find(nb) == G.kmap.end()) {
	  // it->first.toString(tmp); printf("First kmer:  %s\n",tmp);
// 	  it->first.printBinary();
// 	  nb.toString(tmp); printf("Second kmer: %s\n",tmp);
// 	  nb.printBinary(); printf("\n\n");
	  it->second.link &= ~(1<<j);
	}
      }
    }
  }

}

void GraphBuilder::loadBinaryFile(const char *filename) {
  FILE* f = fopen(filename, "rb");
  if (f == NULL) {
    printf("couldn't load file\n");
  } else {
    G.kmap.read_metadata(f);
    G.kmap.read_nopointer_data(f);
    fclose(f);

    Kmer km_del;
    km_del.set_deleted();
    G.kmap.set_deleted_key(km_del);
  }
}

void GraphBuilder::saveBinaryFile(const char *filename) {
  FILE* f = fopen(filename, "wb");
  if (f == NULL) {
    printf("couldn't save file\n");
  } else {
    G.kmap.write_metadata(f);
    G.kmap.write_nopointer_data(f);
    fclose(f);
  }
}


Graph::Graph() {}

Graph::~Graph() {}


void Graph::addKmer(const Kmer &km, const char fw_link, const char bw_link, const bool cov) {
  pair<hmap_t::iterator, bool> ref,ref2;
  ref= kmap.insert(make_pair(km,graph_data()));
  ref2 = kmap.insert(make_pair(km.twin(),graph_data()));

  graph_data &gd = ref.first->second;
  graph_data &gd2 = ref2.first->second;

  if (cov) {
    if (!ref.second) { // this kmer was already there
      gd.cov++;
    }
    
    if (!ref2.second) {
      gd2.cov++;
    }

    size_t fw_ind = char_to_ind(fw_link);
    size_t bw_ind = char_to_ind(bw_link);
    
    if (fw_ind != 0xFF) {
      gd.link |= 1<< fw_ind;
      gd2.link |= 1 << 4+(3-fw_ind); // fw is bw for twin
    }
    
    if (bw_ind != 0xFF) {
      gd.link |= 1 << (4+bw_ind);
      gd2.link |= 1 << (3-bw_ind);
    }
  }  
}

bool Graph::hasKmer(const Kmer &km) const {
  return (kmap.find(km) != kmap.end());
}

size_t Graph::coverage(const Kmer &km) const {
  hmap_t::const_iterator it = kmap.find(km);
  if (it != kmap.end()) {
    return (size_t)it->second.cov;
  } else {
    return 0;
  }
}


size_t Graph::size() const {
  return kmap.size(); 
}

size_t Graph::bucket_count() const {
  return kmap.bucket_count();
}

size_t Graph::deg(const Kmer &km) const {
  return fwDeg(km)+bwDeg(km); 
}

size_t Graph::deg(const_iterator it) const {
  return fwDeg(it)+bwDeg(it);
}

size_t Graph::fwDeg(const Kmer &km) const {
  return help_deg(getLink(km),true);
}

size_t Graph::fwDeg(const_iterator it) const {
  return help_deg(it->second.link, true);
}

size_t Graph::bwDeg(const Kmer &km) const {
  return help_deg(getLink(km),false);
}

size_t Graph::bwDeg(const_iterator it) const {
  return help_deg(it->second.link,false);
}



// edge iterators
EdgeIterator Graph::fwBegin(const Kmer km) const {
  return fwBegin(kmap.find(km));
}

EdgeIterator Graph::fwBegin(const_iterator it) const {
  return EdgeIterator(it, kmap.end(), 0, true);
}

EdgeIterator Graph::fwEnd(const Kmer km) const {
  return fwEnd(kmap.find(km));
}

EdgeIterator Graph::fwEnd(const_iterator it) const {
  return EdgeIterator(it,kmap.end(), 4, true);
}

EdgeIterator Graph::bwBegin(const Kmer km) const {
  return bwBegin(kmap.find(km));
}

EdgeIterator Graph::bwBegin(const_iterator it) const {
  return EdgeIterator(it, kmap.end(), 4, false);
}

EdgeIterator Graph::bwEnd(const Kmer km) const {
  return bwEnd(kmap.find(km));
}

EdgeIterator Graph::bwEnd(const_iterator it) const {
  return EdgeIterator(it, kmap.end(), 8, false);
}


size_t help_deg(size_t link, bool fw) {
  if (link >= 256 || link < 0) {
    return 0xFF;
  } else {
    if (fw) {
      return (link & 0x01) + ((link & 0x02)>>1) + ((link & 0x04)>>2) + ((link & 0x08)>>3);
    } else {
      return ((link & 0x10)>>4) + ((link & 0x20)>>5) + ((link & 0x40)>>6) + ((link & 0x80)>>7);
    }
  }
}

size_t Graph::getLink(const Kmer &km) const {
  hmap_t::const_iterator it = kmap.find(km);
  if (it == kmap.end()) {
    return 0x100;
  } else {
    return it->second.link;
  }
}


void Graph::removeKmer(const Kmer &km) {
  hmap_t::iterator it = kmap.find(km);
  removeKmer(it);
}

void Graph::removeKmer(hmap_t::iterator it) {
  size_t k = Kmer::k;
  char tmp[1024];
  it->first.toString(tmp);
  //printf("Removing Kmer: %s, link: %d\n",tmp,(int)it->second.link);

  size_t first_base = char_to_ind(tmp[0]);
  size_t last_base = char_to_ind(tmp[k-1]);
  uint8_t link = it->second.link;
  
  for (int j = 0; j < 8; ++j) {
    if (link & (1 << j)) {
      //printf("link: %d ",j);fflush(stdout);
      Kmer nb = it->first.getLink(j);
      hmap_t::iterator it2 = kmap.find(nb);
      if (it2 != kmap.end()) {
	size_t ind;
	if (j < 4) {
	  ind = 4 + first_base;
	} else {
	  ind = last_base;
	}
	//nb.toString(tmp); printf("fixing neighbor: %s, index: %d\n",tmp, ind);
	it2->second.link &= ~(1<<ind);
      }      
    }
  }
  
  kmap.erase(it); 
}



void Graph::getContig(Kmer km, vector<Kmer> &out) {
  iterator first, last, next, prev;
  out.clear();

  first = kmap.find(km);
  if (first == kmap.end()) {
    return;
  }

  // go backwards to find starting point
  while (bwDeg(first) == 1) {
    prev = kmap.find(*bwBegin(first)); // find previous
    if (fwDeg(prev) != 1) { // if not in contig break
      break;
    }
    first = prev;
  }

  last = first;
  out.push_back(last->first);
  while (fwDeg(last) == 1) {
    next = kmap.find(*fwBegin(last));
    if (bwDeg(next) != 1) {
      break;
    }
    last = next;
    out.push_back(last->first);
  }
  
}

void Graph::pickSimilar(const vector<string>& seqv, vector<vector<KmerMatch> >& outv, SimilarSequenceOptions opt) {
  vector<vector<int> > H(Kmer::k+2*opt.add+1, vector<int>(Kmer::k+1));
  
  hsim_t sim;
  outv.clear();
  outv.resize(seqv.size());

  for (uint32_t j = 0; j < seqv.size(); ++j) {
    const string& seq = seqv[j];

    for (uint32_t i = 0; i <= seq.size()-opt.k; ++i) {
      string sub= seq.substr(i,opt.k);
      hsim_t::iterator sit = sim.find(sub);
      if (sit == sim.end()) {
	// store hash of opt.k length string to save on space
	pair<hsim_t::iterator ,bool> ret =  sim.insert(make_pair(sub,slist<pair<uint32_t,uint32_t> >()));
	sit = ret.first;
      }
      // push front, since this is a singly linked list (seq_id, seq_pos)
      sit->second.push_front(make_pair(j,i));
    }
  }

  printf("Done building hash table for %zu strings\n",seqv.size());
  printf("Total number of strings hashed: %zu\n",sim.size()); fflush(stdout);
  char tmp[1024];
  // pos stores hits of (seq_id, kmerpos, seq_pos)
  vector<pair<int, pair<int, int> > > pos;
  pos.reserve(10000);
  set<int> gene_matches;

  hsim_t::iterator sim_end = sim.end();

  for (iterator it = kmap.begin(); it != kmap.end(); ++it) {
    it->first.toString(tmp);
    hsim_t::iterator sit;
    // for each substring of the Kmer
    for (uint32_t j = 0; j <= Kmer::k-opt.k; ++j) {
      sit = sim.find(string(tmp+j,opt.k));
      if (sit != sim_end) {
	// we get a hit
	for (slist<pair<uint32_t,uint32_t> >::iterator vit = sit->second.begin(); vit != sit->second.end(); ++vit) {
	  // push back (seq_id, (j, seq_pos))
	  pos.push_back(make_pair(vit->first, make_pair(j,vit->second)));
	}
      }
    }
    
    sort(pos.begin(), pos.end(), PairPairSortSecond()); // sorts by seq_id first, then seq_pos and then j

    gene_matches.clear();

    // for every (gene_id, (kmer_pos,gene_pos))
    for (vector<pair<int, pair<int,int> > >::iterator vit = pos.begin(); vit != pos.end(); ++vit) {
      if (gene_matches.find(vit->first) == gene_matches.end()) {
	const string&  seq = seqv[vit->first];

	int a_beg = vit->second.second - vit->second.first - opt.add;
	a_beg = (a_beg >= 0) ? a_beg : 0;
	int a_end = vit->second.second + Kmer::k - vit->second.first + opt.add;
	a_end = (a_end < 0 || size_t(a_end) <= seq.size()) ? a_end : seq.size();
	string a;
	try {
	  a = seq.substr(a_beg, a_end-a_beg);
	} catch (std::out_of_range& e) {
	  printf("out of range, seq len: %zu, beg: %d, end: %d, size:%d\n", seq.size(), a_beg, a_end, a_end-a_beg);
	  continue;
	}
	
	string b(tmp);

	// now compare the two strings
	int M = a_end - a_beg;
	int N = Kmer::k;
	for (int i = 0; i <= M; ++i) {
	  H[i][0] = i;
	}
	for (int j = 0; j <= N; ++j) {
	  H[0][j] = j;
	}
	int w;
	for (int i = 1; i <= M; ++i) {
	  for (int j = 1; j <= N; ++j) {
	    w = 1; // mismatch is 1
	    if (a[i]==b[j]) {
	      w = 0; // match is 1
	    }
	    H[i][j] = min(H[i-1][j-1]+w ,min(H[i-1][j]+1,H[i][j-1]+1)); // take the min of all entries, 1 for gaps
	  }
	}
	
	if ( H[M][N]  <= (opt.dist + 2*opt.add)) {
	  // if we found a match, push back to the appropriate gene
	  int refpos = vit->second.second - vit->second.first;
	  gene_matches.insert(vit->first);
	  refpos = (refpos >= 0) ? refpos : 0;
	  outv[vit->first].push_back(KmerMatch(it->first, (uint32_t) refpos));
	}
      } 
    }
    
    // clear the pos vector
    pos.clear();
  }

}

void Graph::clipTips(ClipTipOptions opt) {
  printf("clipTips called, size before: %zu ..",size());
  bool modified = true;
  //printDot();

  while (modified) {
    modified = false;
    
    iterator it;
    for (it = kmap.begin(); it != kmap.end(); ++it) {
      if (bwDeg(it) == 0) {
	modified = clipTip(it,opt);
      }
    }
    
    if (modified) {
      deleteMarked();
      //printDot();
    }
  }
  printf("..after: %zu\n",size());
}

bool Graph::clipTip(iterator first, ClipTipOptions opt) {
  bool clipped = false;
  iterator last = first;
  iterator next = kmap.end();
  
  int tipLength = 1;
  
  while (fwDeg(last) == 1) {
    edge_iterator e = fwBegin(last);  // find first and only link
    if (e == fwEnd(last)) {
      // error something is wrong, should print to cerr, we should build an 
      // error reporting mechanism latere
      return false;
    }
    
    next = kmap.find(*e);
    if (next == kmap.end()) {
      return false;
    }
    if (bwDeg(next) == 1) {
      last = next;
      ++tipLength;
    } else {
      break;
    }
  }

  //printf("Found tip of length %d\n",tipLength);

  

  // now first .. last is a tip, check if we should remove or not
  if (tipLength > opt.maxLength) {
    return false;
  }
  
  if (fwDeg(last) == 0) {
    clipped = true;
  } else if (fwDeg(last) > 1) {
    clipped = false;
  } else {
    // next is the end of the tip and will have bwDeg > 1;
    size_t max_cov = 0;
    
    for(edge_iterator e = bwBegin(next); e != bwEnd(next); ++e) {
      iterator prev = kmap.find(*e);
      if (prev != kmap.end()) {
	max_cov = (max_cov > prev->second.cov) ? max_cov : prev->second.cov;
      }
    }
    if (last->second.cov < max_cov) {
      clipped = true;
    }
  }

  if (clipped) {
    // traverse the tip again
    for (iterator tip = first; tip != last && tip != kmap.end();) {
      tip->second.deleted = 1;
      tip = kmap.find(*fwBegin(tip));
    }
    last->second.deleted = 1;
  }
  return clipped;
}

void Graph::deleteMarked() {
  iterator it,cur;
  for (it = kmap.begin(); it != kmap.end(); ) {
    cur = it;
    ++it;
    cur->second.visited = 0; // clear the visited flag
    if (cur->second.deleted != 0) {
      iterator tw_it = kmap.find(cur->first.twin());
      if (it == tw_it && it != kmap.end()) {
	++it;
      }

      removeKmer(cur);
      if (tw_it != kmap.end()) {
	removeKmer(tw_it);
      }
    } 
  }
}



void Graph::printDot() {
  printf("DOT FILE\n\n");
  printf("digraph G{\ngraph [rankdir=LR, fontname=\"Courier\"];\nnode [shape=record];\n");
  for (iterator it = kmap.begin(); it != kmap.end(); ++it) {
    char tmp[1024];
    it->first.toString(tmp);
    printf("%s;\n",tmp);
  }
  printf("\n");
  for (iterator it = kmap.begin(); it != kmap.end(); ++it) {
    char tmp1[1024];
    char tmp2[1024];
    it->first.toString(tmp1);
    for(edge_iterator e = fwBegin(it); e != fwEnd(it); ++it) {
      e->toString(tmp2);
      printf("%s -> %s;\n",tmp1,tmp2);
    }
  }
  printf("}\n");
}


hmap_t::iterator Graph::begin() { return kmap.begin(); }
hmap_t::iterator Graph::end() { return kmap.end(); }

hmap_t::const_iterator Graph::begin() const { return kmap.begin(); }
hmap_t::const_iterator Graph::end() const { return kmap.end(); }

void Graph::testGraph() {
  hmap_t::iterator it,it2;

  size_t cnt = 0;
  printf("testing twins\n");
  for (it = kmap.begin(); it != kmap.end(); ++it) {
    Kmer tw = it->first.twin();
    ++cnt;
    if (kmap.find(tw) == kmap.end()) {
      char tmp[1024];
      printf("could not find twin in map\n");
      it->first.toString(tmp); printf("kmer  %s\n",tmp);
      it->first.printBinary();
      tw.toString(tmp); printf("twin  %s\n",tmp);
      tw.printBinary();
      return;
    }
  }
  printf("tested %zu things\n",cnt);


  cnt = 0;
  printf("testing neighbor structure\n");
  for (it = kmap.begin(); it != kmap.end(); ++it) {
    size_t link = it->second.link;
    
    for (int j = 0; j < 8; ++j) {
      if (link & (1 << j)) {
	++cnt;
	Kmer km = it->first.getLink(j);
	//char tmp[1024];
	char tmp1[1024];
	char tmp2[1024];
	it->first.toString(tmp1);
	km.toString(tmp2);
	it2 = kmap.find(km);
	if (it2 == kmap.end()) {
	  printf("could not find neighbor in map\n");

	  printf("First kmer: %s\n",tmp1);
	  it->first.printBinary();
	  Kmer tw = it->first.twin();
	  if (tw < it->first) {
	    printf("not a representative\n");
	  }
	  printf("Second kmer: %s\n",tmp2);
	  km.printBinary();
	  printf("Link information %zu, index %d\n",link,j);
	  
	  return;
				
	} else {
	  // this is what we expect to find in the other kmer
	  char char_to_check = tmp1[(j<4) ? 0 : (Kmer::k-1)];
	  uint8_t ind_to_check = ((j<4) ? 4: 0)+char_to_ind(char_to_check);
	  
	  if (!(it2->second.link & (1 << ind_to_check))) {
	    printf("neighbor exists, does not point back\n");
	    printf("checked char %c\n", char_to_check);
	    printf("looked at index %d\n",ind_to_check);

	    printf("First kmer: %s\n",tmp1);
	    it->first.printBinary();
	    //it->first.twin().toString(tmp);
	    //printf("It's twin:  %s\n",tmp);

	    printf("Second kmer: %s\n",tmp2);
	    km.printBinary();
	    //km.twin().toString(tmp);
	    //printf("It's twin:  %s\n",tmp);
	    printf("Link information for first kmer %zu, index %d, coverage: %d\n",link,j,it->second.cov);
	    printf("Link information for second kmer %d, index %d, coverage: %d\n",it2->second.link,ind_to_check, it2->second.cov);
	    printf("tested %zu things\n",cnt);
	    return;
	  }
	}
      }
    }
  }

  printf("tested %zu things\n",cnt);

  /*  for (it = kmap.begin(); it != kmap.end(); ++it) {
    size_t link = it->second.link;
    for (int j = 0; j < 8; ++j) {
      Kmer km = it->first.getLink(j);
      Kmer tw = km.twin();
      Kmer otw = it->first.twin();

    }
    }*/
}

EdgeIterator::EdgeIterator(hmap_t::const_iterator  _v, hmap_t::const_iterator _v_end, 
			   const size_t _pos, const bool _fw) : v(_v), v_end(_v_end), pos(_pos), fw(_fw) {
  while (!pos_valid()) {
    ++pos;
    if (v == v_end || (fw && pos >= 4) || (!fw && pos >=8)) {
      pos = fw ? 4 : 8;
      break;
    }
  }
  if (v != v_end && pos_valid()) {
    cur = v->first.getLink(pos);
  }
}

bool EdgeIterator::pos_valid() const {
  return ((fw && pos >=0 && pos < 4) || (!fw && pos >=4 && pos < 8) ) && v != v_end &&(v->second.link & (1<<pos));
}

Kmer& EdgeIterator::operator*() {
  assert(pos_valid());
  return cur;
}

Kmer* EdgeIterator::operator->() {
  return &(this->operator*());
}

EdgeIterator& EdgeIterator::operator++() { 
  ++pos;
  while (v != v_end && !(v->second.link & (1<<pos))) {
    ++pos;
    if (v == v_end || (fw && pos >= 4) || (!fw && pos >= 8)) {
      pos = fw ? 4 : 8;
      break;
    }
  }
  if (v != v_end && pos_valid()) {
    cur = v->first.getLink(pos);
  }
  return *this;
}

EdgeIterator EdgeIterator::operator++(int) {
  EdgeIterator ret(*this);
  this->operator++();
  return ret;
}

bool EdgeIterator::operator==(const EdgeIterator& o) const {
  return v == o.v && v_end == o.v_end && fw == o.fw && pos == o.pos;
}

bool EdgeIterator::operator!=(const EdgeIterator& o) const {
  return !operator==(o);
}
