#ifndef GRAPH_H
#define GRAPH_H

#include <google/sparse_hash_map>
#include <google/sparse_hash_set>

#include <queue>
#include <numeric>
#include <vector>
#include <ext/slist>

#include "fastq.hpp"
#include "kmer.hpp"
#include "bloom_filter.hpp"
#include "hash.h"

using namespace std;
using __gnu_cxx::slist;
using google::sparse_hash_map;
using google::sparse_hash_set;


/** Useful hash functions **/
struct KmerHash {
  size_t operator()(const Kmer &km) const {
    return km.hash();
  }
};

struct StringHash { 
  size_t operator()(const string& x) const {
    return SuperFastHash(x.c_str(), x.size());
  }
};

/** information for k-mers **/
class graph_data {
public:
  graph_data() : cov(0),visited(0),deleted(0),link(0) {}
  uint32_t cov:22;
  uint32_t visited:1;
  uint32_t deleted:1;
  uint8_t link;

  size_t coverage() const {
    return (size_t) cov;
  }
  
  bool isVisited() const {
    return (bool) visited;
  }

  bool isDeleted() const {
    return (bool) deleted;
  }

  bool operator==(const graph_data& o) const{
    return (cov==o.cov) && (link  == o.link);
  }
};


/** option helper classes **/
class ClipTipOptions {
public:
  ClipTipOptions() : maxLength(Kmer::k) {}
  
  int maxLength;
};

class SimilarSequenceOptions {
public:
  SimilarSequenceOptions() : k(Kmer::k) , dist(4) , add(2) {}

  int add;
  int dist;
  int k;
};

struct KmerMatch {
  KmerMatch(const Kmer &_km, const uint32_t &_refpos) : km(_km), refpos(_refpos) {}
  KmerMatch() : refpos(0xFFFFFFFF) {}
  Kmer km;
  uint32_t refpos;
};

// useful typedefs
typedef sparse_hash_map<Kmer, graph_data, KmerHash> hmap_t;
typedef sparse_hash_map<Kmer, pair<const Kmer,int>, KmerHash> prevMap_t;
typedef sparse_hash_map<string, slist<pair<uint32_t,uint32_t> >, StringHash > hsim_t;
typedef sparse_hash_set<Kmer, KmerHash> hset_t;
typedef sparse_hash_set<string, StringHash> hstringset_t;


class Graph;


/** Helper class that builds the graph from a set of reads **/

class GraphBuilder {
 public:
  // exp_numer_of_reads is a hint used to create a reasonably sized bloom filter
  GraphBuilder(Graph &_G, size_t exp_number_of_reads = 15000000) : G(_G),exp_size(exp_number_of_reads), f(NULL) {}
  ~GraphBuilder();

  void processFastqFile(FastqFile &F);

  void loadBinaryFile(const char *filename);
  void saveBinaryFile(const char *filename);

 private:
  
  Graph &G;
  size_t exp_size;
  bloom_filter *f;
};

/** iterator decoration **/
class EdgeIterator {
public:
  // constructor
  EdgeIterator(hmap_t::const_iterator  _v, hmap_t::const_iterator _v_end, const size_t _pos, const bool _fw);
  
  Kmer& operator*();
  Kmer* operator->();
  
  EdgeIterator& operator++();
  EdgeIterator operator++(int);
  
  bool operator==(const EdgeIterator& o) const;
  bool operator!=(const EdgeIterator& o) const;
  bool pos_valid() const;

private:
  hmap_t::const_iterator v,v_end;
  size_t pos;
  bool fw;
  Kmer cur;
};



/** The main graph class **/
class Graph {
 public:
  Graph();
  ~Graph();

  // friend classes
  friend class GraphBuilder;
  friend class EdgeIterator;

  typedef hmap_t::iterator iterator;
  typedef hmap_t::const_iterator const_iterator;
  typedef EdgeIterator edge_iterator;

  // add a new kmer to the graph with fw and bw links
  // links should be one of AGCT or \0 (if no info)
  void addKmer(const Kmer &km, const char fw_link = '\0', const char bw_link = '\0', const bool cov = false);
  
  // removes the kmer from the graph and updates all
  // link information for neighboring kmers
  void removeKmer(const Kmer &km);
  void removeKmer(iterator it);
  
  // returns true if the kmer is in the graph
  bool hasKmer(const Kmer &km) const;


  // iterator support
  iterator begin();
  iterator end();
  
  const_iterator begin() const;
  const_iterator end() const;

  EdgeIterator fwBegin(const Kmer) const;
  EdgeIterator fwBegin(const_iterator it) const;
  EdgeIterator fwEnd(const Kmer) const;
  EdgeIterator fwEnd(const_iterator it) const;
  
  EdgeIterator bwBegin(const Kmer) const;
  EdgeIterator bwBegin(const_iterator it) const;
  EdgeIterator bwEnd(const Kmer) const;
  EdgeIterator bwEnd(const_iterator it) const;
  

  // returns the coverage of the kmer or 0 if the kmer doesn't exist
  size_t coverage(const Kmer &km) const;
  
  // do we really need this?
  size_t getLink(const Kmer &km) const;

  // fwDeg returns the number of forward neighbors
  // bwDeg -                   - backward neighbors
  // deg = fwDeg+bwDeg
  size_t deg(const Kmer &km) const;
  size_t deg(const_iterator it) const;
  size_t fwDeg(const Kmer &km) const;
  size_t fwDeg(const_iterator it) const;
  size_t bwDeg(const Kmer &km) const;
  size_t bwDeg(const_iterator it) const;

  size_t size() const; 
  size_t bucket_count() const;


  /* ---- graph operations ---- */

  void testGraph();
  void printDot();

  void getContig(Kmer km, vector<Kmer> &out);
  void pickSimilar(const vector<string>& seqv, vector<vector<KmerMatch> >& outv, SimilarSequenceOptions opt = SimilarSequenceOptions());

  // graph modifications
  void clipTips(ClipTipOptions opt = ClipTipOptions());
  void deleteMarked();


 private:
  bool clipTip(iterator it, ClipTipOptions opt);
  hmap_t kmap;
};

#endif
